/*
  MemoryGame.cpp - library used to play the Memory Game project.
  Created by Stefan Alexandru Apostol, May 29, 2023.
*/

#include "Arduino.h"
#include "MemoryGame.h"

// Default constructor
MemoryGame::MemoryGame()
{
  g_sleepPeriod = 0;
  _score = 0;
  _highScore = 0;
  _level = 0;
  _lives = 3;
  _greenValue = 255;
  _lettersRead = 0;
  _keyPressed = 0;
}

// Default destructor
MemoryGame::~MemoryGame()
{
  delete _lcd;
  delete _rtc;
  delete _time;
}

void MemoryGame::init()
{
  // Use Serial for debug messages
  Serial.begin(9600);
  
  initializeMenuButtons();
  initializeRGBLed();
  initializeBuzzer();
  initializeLCD();
  initializeRTC();
  initializeSensors();
  initializeSDCard();

  // Setup timer
  cli();  
  configure_timer1();
  init_timer1();
  // Activate global interrupts
  sei();

  _state = SLEEP;
}

void MemoryGame::draw()
{
  if (_state == SLEEP)
    drawStandBy();

  if (_state == MENU)
    drawMenu();

  if (_state == RUNNING || _state == WAITING || _state == REPEAT)
    drawRunningGame();

  if (_state == GAME_OVER)
    drawGameOverScreen();

  if (_state == TIME_SHOW)
    drawTime();

  if (_state == SCORE_SHOW)
    drawScore();
}

void MemoryGame::update()
{
  // Make sure to exit from the time setting after 5 seconds
  if (_state == TIME_SHOW && g_sleepPeriod == 10) {
    _lcd->clear();
    g_sleepPeriod = 0;

    _state = MENU;
    return;
  }

  // Make sure to exit from the highscore setting after 5 seconds
  if (_state == SCORE_SHOW && g_sleepPeriod == 10) {
    _lcd->clear();
    g_sleepPeriod = 0;

    _state = MENU;
    return;
  }

  // In the RUNNING state of the game generate a sequence of characters
  // and go in WAITING state
  if (_state == RUNNING) {
    generateSequence();
    _lcd->clear();

    _state = WAITING;
    return;
  }

  // In the REPEAT state of the game show again the sequence generated
  // and go in WAITING state
  if (_state == REPEAT) {
    repeatSequence();
    _lcd->clear();

    _state = WAITING;
    return;
  }
}

void MemoryGame::input()
{

  // Let the user be able to input characters on the analog keyboard
  // when the game is in the WAITING state.
  if (_state == WAITING)
    readInput();

  readLeftButton();
  readMiddleButton();
  readRightButton();
}

void MemoryGame::initializeMenuButtons()
{
  pinMode(LEFT, INPUT_PULLUP);
  pinMode(MIDDLE, INPUT_PULLUP);
  pinMode(RIGHT, INPUT_PULLUP);
}

void MemoryGame::initializeRGBLed()
{
  pinMode(PIN_RED, OUTPUT);
  pinMode(PIN_GREEN, OUTPUT);
  pinMode(PIN_BLUE, OUTPUT);
}

void MemoryGame::initializeBuzzer()
{
  pinMode(BUZZER, OUTPUT);
}

void MemoryGame::initializeLCD()
{
  _lcd = new LiquidCrystal_I2C(LCD, 16, 2);
  _lcd->init();  
  _lcd->backlight();

    // Heart character
  byte customChar[8] = {
    0b00000,
    0b01010,
    0b11111,
    0b11111,
    0b01110,
    0b00100,
    0b00000,
    0b00000
  };
  _lcd->createChar(0, customChar);
}

void MemoryGame::initializeRTC()
{
  _rtc = new DS3231();
  _time = new Time();
}

void MemoryGame::initializeSensors()
{
  _ds = new DS18B20(TEMP);
}

void MemoryGame::initializeSDCard()
{
  if (!SD.begin(CS)) {
    Serial.println("SD CARD FAILED, OR NOT PRESENT!");
    while (1); // don't do anything more
  }

  // Create log.txt file if it doesn't exist
  File file = SD.open("log.txt", FILE_READ);
  if (file) {
    Serial.println("SD Card: log.txt file opened with success");
    file.close();
  } else {
    Serial.println("SD Card: log.txt file failed to open");
    file = SD.open("log.txt", FILE_WRITE);
    file.close();
  }

  // Create highscore.txt file if it doesn't exist
  file = SD.open("score.txt", FILE_READ);
  if (file) {
    Serial.println("SD Card: score.txt file opened with success");
    file.close();
  } else {
    Serial.println("SD Card: score.txt file failed to open");
    file = SD.open("score.txt", FILE_WRITE);
    file.println(0);
    file.close();
  }
}

void MemoryGame::drawStandBy()
{
  for (uint8_t i = 0; i < g_sleepPeriod; i++) {
    _lcd->setCursor(i % 16, i / 16);
    _lcd->print("z");
  }

  if (g_sleepPeriod == 33) {
    _lcd->clear();
    g_sleepPeriod = 0;
  }

  analogWrite(PIN_GREEN, _greenValue);
}

void MemoryGame::drawMenu()
{
  _lcd->setCursor(_currSetting / 2 * 7, _currSetting % 2);
  _lcd->print(">");
  _lcd->setCursor(1, 0);
  _lcd->print("Start");
  _lcd->setCursor(1, 1);
  _lcd->print("Exit");
  _lcd->setCursor(8, 0);
  _lcd->print("ShowTime");
  _lcd->setCursor(8, 1);
  _lcd->print("HghScore");
}

void MemoryGame::drawTime()
{
  _time->SetValues(_rtc);
  _lcd->setCursor(0, 0);
  _lcd->print(_time->GetTimeMessage());
  _lcd->setCursor(0, 1);
  _lcd->print(_time->GetDayMessage());
}

void MemoryGame::drawScore()
{
  _lcd->setCursor(0, 0);
  _lcd->print("HighScore:");
  _lcd->setCursor(0, 1);
  _lcd->print(_highScore);
}

void MemoryGame::drawRunningGame()
{
  String s_level = "Level ";
  s_level += _level;
  _lcd->setCursor(0, 0);
  _lcd->print(s_level);

  // Draw hearts
  for (int i = 0; i < _lives; i++) {
    _lcd->setCursor(i, 1);
    _lcd->write((byte)0);
  }

  // Draw score
  String scoreMsg = "S:";
  if (_score < 10)
    scoreMsg += " ";
  scoreMsg += _score;
  _lcd->setCursor(12, 0);
  _lcd->print(scoreMsg);

  // Draw highscore
  String hScoreMsg = "H:";
  if (_highScore < 10)
    hScoreMsg += " ";
  hScoreMsg += _highScore;
  _lcd->setCursor(12, 1);
  _lcd->print(hScoreMsg);
}

void MemoryGame::drawGameOverScreen()
{
  _lcd->setCursor(3 + _currRetryOption * 6, 1);
  _lcd->print(">");

  _lcd->setCursor(5, 0);
  _lcd->print("Retry?");

  _lcd->setCursor(4, 1);
  _lcd->print("Yes");

  _lcd->setCursor(10, 1);
  _lcd->print("No");
}

void MemoryGame::readInput()
{
  // When all the characters have been guessed correctly move
  // to the next level and light the RGB led to GREEN;
  if (_lettersRead == 3 + (_level / 3)) {
    _level += 1;
    _score += 2;
    if (_score > _highScore)
      _highScore = _score;
    _lettersRead = 0;
    setRGBLed(0, 255, 0);
    delay(1000);
    setRGBLed(0, 0, 0);

    _state = RUNNING;
    return;
  }

  // Get the letter from the analog button that was pressed
  char letter = decodeInput(analogRead(KEYBOARD));
  if (letter == 'Z') {
    _keyPressed = millis();
  }

  // Here make sure that input is taken only if the key is pressed for
  // at least 100 miliseconds
  if (millis() - _keyPressed >= 100) {
    String message = "Analog key ";
    message += letter;
    message += " was pressed";
    printLogMessage(message.c_str());
    _lcd->setCursor(7, 1);
    _lcd->print(letter);
    _keyPressed = millis();
    setTone(letter, 100);
    delay(500);

    // If the character was not guessed sub a life and make the
    // game repeat the sequence
    if (letter != _buffer[_lettersRead]) {
      _lives -= 1;
      _lettersRead = 0;
      setRGBLed(255, 0, 0);
      delay(1000);
      setRGBLed(0, 0, 0);
      _lcd->clear();

      // When the player has no more lives left go into GAME_OVER state
      if (_lives == 0) {
        _state = GAME_OVER;
        return;
      }

      _state = REPEAT;
      return;
    }

    _lcd->clear();
    _lettersRead += 1;
  }
}

void MemoryGame::setRGBLed(uint8_t r, uint8_t g, uint8_t b)
{
  analogWrite(PIN_RED, r);
  analogWrite(PIN_GREEN, g);
  analogWrite(PIN_BLUE,  b);
}

void MemoryGame::setTone(char input, unsigned long duration)
{
  if (input == 'A') {
    tone(BUZZER, NOTE_A2, duration);
    return;
  }

  if (input == 'B') {
    tone(BUZZER, NOTE_B2, duration);
    return;
  }

  if (input == 'C') {
    tone(BUZZER, NOTE_C2, duration);
    return;
  }

  if (input == 'D') {
    tone(BUZZER, NOTE_D2, duration);
    return;
  }

  if (input == 'E') {
    tone(BUZZER, NOTE_E2, duration);
    return;
  }

  if (input == 'F') {
    tone(BUZZER, NOTE_F2, duration);
    return;
  }
}

void MemoryGame::readLeftButton()
{
  const int button = LEFT;

  // read the state of the switch into a local variable:
  int reading = digitalRead(button);

  // check to see if you just pressed the button
  // (i.e. the input went from LOW to HIGH), and you've waited long enough
  // since the last press to ignore any noise:

  // If the switch changed, due to noise or pressing:
  if (reading != _lastButtonState[button - LEFT]) {
    // reset the debouncing timer
    _lastDebounceTime[button - LEFT] = millis();
  }
    
  //Serial.println(button);

  if ((millis() - _lastDebounceTime[button - LEFT]) > _debounceDelay[button - LEFT]) {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:

    // if the button state has changed:
    if (reading != _buttonState[button - LEFT]) {
      _buttonState[button - LEFT] = reading;

      // only toggle the LED if the new button state is HIGH
      if (
        _buttonState[button - LEFT] == HIGH 
        && _state == MENU) {
          if (_currSetting == 0)
            _currSetting = 3;
          else
            _currSetting = _currSetting - 1;

          _lcd->clear();
          return;
      }

      if (
        _buttonState[button - LEFT] == HIGH 
        && _state == GAME_OVER) {
          if (_currRetryOption == 0)
            _currRetryOption = 1;
          else
            _currRetryOption = _currRetryOption - 1;

          _lcd->clear();
          return;
      }
    }
  }

  _lastButtonState[button - LEFT] = reading;
}

void MemoryGame::readMiddleButton()
{
  const int button = MIDDLE;

  // read the state of the switch into a local variable:
  int reading = digitalRead(button);

  // check to see if you just pressed the button
  // (i.e. the input went from LOW to HIGH), and you've waited long enough
  // since the last press to ignore any noise:
  // If the switch changed, due to noise or pressing:
  if (reading != _lastButtonState[button - LEFT]) {
    // reset the debouncing timer
    _lastDebounceTime[button - LEFT] = millis();
  }
    
  //Serial.println(button);

  if ((millis() - _lastDebounceTime[button - LEFT]) > _debounceDelay[button - LEFT]) {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:

    // if the button state has changed:
    if (reading != _buttonState[button - LEFT]) {
      _buttonState[button - LEFT] = reading;
      // only toggle the LED if the new button state is HIGH
      if (_buttonState[button - LEFT] == HIGH) {
        if (_state == SLEEP) {
          _lcd->clear();

          _state = MENU;
        } else if (_state == MENU) {
          selectSetting();
        } else if (_state == GAME_OVER) {
          switch(_currRetryOption) {
            case 0:
              _level = 0;
              _lives = 3;
              _score = 0;
              _currRetryOption = 0;
              _lcd->clear();

              _state = RUNNING;
              return;
            case 1:
              _level = 0;
              _lives = 3;
              _score = 0;
              _currRetryOption = 0;
              _lcd->clear();

              _state = MENU;
              return;
          }
        }
      }
    }
  }

  _lastButtonState[button - LEFT] = reading;
}

void MemoryGame::readRightButton()
{
  const int button = RIGHT;

  // read the state of the switch into a local variable:
  int reading = digitalRead(button);

  // check to see if you just pressed the button
  // (i.e. the input went from LOW to HIGH), and you've waited long enough
  // since the last press to ignore any noise:

  // If the switch changed, due to noise or pressing:
  if (reading != _lastButtonState[button - LEFT]) {
    // reset the debouncing timer
    _lastDebounceTime[button - LEFT] = millis();
  }
    
  //Serial.println(button);

  if ((millis() - _lastDebounceTime[button - LEFT]) > _debounceDelay[button - LEFT]) {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:

    // if the button state has changed:
    if (reading != _buttonState[button - LEFT]) {
      _buttonState[button - LEFT] = reading;

      // only toggle the LED if the new button state is HIGH
      if (
        _buttonState[button - LEFT] == HIGH 
        && _state == MENU) {
          _currSetting = (_currSetting + 1) % 4;

          _lcd->clear();
          return;
      }
      if (
        _buttonState[button - LEFT] == HIGH 
        && _state == GAME_OVER) {
          _currRetryOption = (_currRetryOption + 1) % 2;

          _lcd->clear();
          return;
      }
    }
  }

  _lastButtonState[button - LEFT] = reading;
}

void MemoryGame::selectSetting()
{
  // Treat every possible setting and go to the next state of the
  // game based on the setting selected.
  if (_currSetting == 0) {
    _lcd->clear();
    g_sleepPeriod = 0;
    _currSetting = 0;
    setRGBLed(0, 0, 0);

    _state = RUNNING;
    return;
  }

  if (_currSetting == 1) {
    _lcd->clear();
    g_sleepPeriod = 0;
    _currSetting = 0;

    _state = SLEEP;
    return;
  }

  if (_currSetting == 2) {
    _lcd->clear();
    g_sleepPeriod = 0;
    _currSetting = 0;

    _state = TIME_SHOW;
    return;
  }

  if (_currSetting == 3) {
    _lcd->clear();
    g_sleepPeriod = 0;
    _currSetting = 0;

    _state = SCORE_SHOW;
    return;
  }
}

char MemoryGame::decodeInput(int value) {
  if (value >= 1018 && value <= 1023)
    return 'A';

  if (value >= 1010 && value <= 1017)
    return 'B';

  if (value >= 1000 && value <= 1009)
    return 'C';

  if (value >= 951 && value <= 999)
    return 'D';

  if (value >= 900 && value <= 950)
    return 'E';

  if (value >= 400 && value <= 899)
    return 'F';

  return 'Z';
}

void MemoryGame::generateSequence()
{
  for (uint8_t i = 0; i < 10; i++)
      _buffer[i] = '0';

  char chars[6] = {'A', 'B', 'C', 'D', 'E', 'F'};
  uint8_t lettersCount = 3 + _level / 3;
  lettersCount > 10 ? 10 : lettersCount;
  for (uint8_t i = 0; i < lettersCount; i++) {
    _buffer[i] = randLetter(chars);
    while (i > 0 && _buffer[i] == _buffer[i - 1]) {
      _buffer[i] = randLetter(chars);
    }

    _lcd->setCursor(7, 1);
    _lcd->print(_buffer[i]);
    setTone(_buffer[i], 100);
    delay(1000 - 100 * (_level / 3));
  }
}

void MemoryGame::repeatSequence()
{
  uint8_t lettersCount = 3 + _level / 3;
  lettersCount > 10 ? 10 : lettersCount;
  for (uint8_t i = 0; i < lettersCount; i++) {
    _lcd->setCursor(7, 1);
    _lcd->print(_buffer[i]);
    setTone(_buffer[i], 100);
    delay(1000 - 100 * (_level / 3));
  }
}

char MemoryGame::randLetter(char *chars)
{
  int temp = (int)floor(_ds->getTempC() * 100);
  temp = temp % 1000 + temp / 10 % 10 + temp / 100 % 10 + temp / 1000;
  for (int i = 0; i < temp; i++) {
    int x = random() % 6;
    int y = random() % 6;
    char aux;
    aux = chars[x];
    chars[x] = chars[y];
    chars[y] = aux;
  }

  return chars[random() % 6];
}

void MemoryGame::printLogMessage(char *message)
{
  _time->SetValues(_rtc);
  String logMessage = _time->GetTimeMessage();
  logMessage += " - ";
  logMessage += String(message);
  Serial.println(logMessage);
  File file = SD.open("log.txt", FILE_WRITE);
  if (file) {
    Serial.println("Writting to SD card");
    file.println(logMessage);
    file.close();
  }
}

void MemoryGame::configure_timer1() {
  // Interrupt wiht 2Hz frequence and CTC mode
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;
  OCR1A = 31249;            // compare match register 16MHz/256/2Hz-1
  TCCR1B |= (1 << WGM12);   // CTC mode
  TCCR1B |= (1 << CS12);    // 256 prescaler 
}
 
void MemoryGame::init_timer1() {
  TIMSK1 |= (1 << OCIE1A);  // enable timer compare interrupt
}
