#ifndef MEMORY_GAME_H
#define MEMORY_GAME_H

/*
  MemoryGame.h - library used to play the Memory Game project.
  Created by Stefan Alexandru Apostol, May 29, 2023.
*/

#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <DS3231.h>
#include <DS18B20.h>
#include <SPI.h>
#include <SD.h>
#include "Arduino.h"

#define NOTE_A2  600
#define NOTE_B2  650
#define NOTE_C2  700
#define NOTE_D2  750
#define NOTE_E2  800
#define NOTE_F2  850

class MemoryGame {
public:
  MemoryGame();
  ~MemoryGame();

  /*
  * This function sets the private game variables and initializes
  * all the game modules that will be used
  */
  void init();

  /*
  * This function will draw on the LCD the state of the game 
  */
  void draw();

  /*
  * This function will update some variables based on the user
  * input or the game state
  */
  void update();

  /*
  * This function will consider the user input and based on it the
  * game will change it's state.
  */
  void input();

  // Getters
  uint8_t   getGreenValue()               { return _greenValue; }

  // Setters
  void      setGreenValue(uint8_t val)   { _greenValue = val; }

  uint8_t   g_sleepPeriod; // Variable used to count internal period of 0.5 seconds

  // Led Constants
  const int PIN_RED = 6;
  const int PIN_GREEN = 5;
  const int PIN_BLUE = 3;

  // Buzzer constant
  const int BUZZER = 2;

  // Menu buttons constants
  const int LEFT = 7;
  const int MIDDLE = 8;
  const int RIGHT = 9;

  // LCD constant
  const int LCD = 0x27;

  // Temperature sensor constant
  const int TEMP = 4;

  // Light sensor constant
  const int LIGHT = A1;

  // Analog keyboard constant
  const int KEYBOARD = A2;

  // SD card constants
  const int CS = 10;
  const int SCK = 13;
  const int MOSI = 11;
  const int MISO = 12;

private:
  /*
  * This is an enum structure that stores all the game's possible states
  * that it can go through. 
  */
  enum GameState {
    SLEEP,
    MENU,
    TIME_SHOW,
    SCORE_SHOW,
    RUNNING,
    WAITING,
    REPEAT,
    GAME_OVER
  };

  /*
  * This is a structure used to store data related to time 
  */
  struct Time {
    byte hour;
    byte minute;
    byte second;
    byte year;
    byte month;
    byte day;
    bool h12;
    bool pmTime;
    bool century;

    // Default constructor
    Time()
    {
      hour = 0;
      minute = 0;
      second = 0;
      year = 0;
      month = 0;
      day = 0;
      h12 = 0;
      pmTime = 0;
      century = 0;
    }

    // This function sets the data related to the current time and date
    void SetValues(const DS3231 *rtc)
    {
      hour = rtc->getHour(h12, pmTime);
      minute = rtc->getMinute();
      second = rtc->getSecond();
      year = rtc->getYear();
      month = rtc->getMonth(century);
      day = rtc->getDate();
    }

    // This function get a relevant message that tells the current time.
    String GetTimeMessage()
    {
      String message = "Time : ";
      if (hour < 0b1010)
        message += "0";
      message += hour;
      message += ":";
      if (minute < 0b1010)
        message += "0";
      message += minute;
      message += ":";
      if (second < 0b1010)
        message += "0";
      message += second;

      return message;
    }

    // This function get a relevant message that tells the current date.
    String GetDayMessage()
    {
      String message = "Date : ";
      if (day < 0b1010)
        message += "0";
      message += day;
      message += "/";
      if (month < 0b1010)
        message += "0";
      message += month;
      message += "/";
      if (year < 0b1010)
        message += "0";
      message += year;

      return message;
    }
  };

  /*
  * This function is used to initialize the digital buttons 
  */
  void initializeMenuButtons();
  /*
  * This function is used to initialize the RGB led 
  */
  void initializeRGBLed();
  /*
  * This function is used to initialize the buzzer 
  */
  void initializeBuzzer();
  /*
  * This function is used to initialize the LCD display
  */
  void initializeLCD();
  /*
  * This function is used to initialize the RTC module 
  */
  void initializeRTC();
  /*
  * This function is used to initialize the light sensor
  * and the temperature sensor used. The temperature sensor
  * communicates on digital signal, while the light sensor
  * communicates on analog signal 
  */
  void initializeSensors();
  /*
  * This function is used to initialize the Mini SD card
  */
  void initializeSDCard();

  /*
  * This function is used to draw the sleep scene of the game 
  */
  void drawStandBy();
  /*
  * This function is used to draw the settings menu of the game 
  */
  void drawMenu();
  /*
  * This function is used to draw the time and date messages on
  * the LCD 
  */
  void drawTime();
  /*
  * This function is used to draw the highscore on the LCD
  */
  void drawScore();
  /*
  * This function is used to draw the background of the running game
  */
  void drawRunningGame();
  /*
  * This function is used to draw the game over scene of the game 
  */
  void drawGameOverScreen();

  /*
  * This function is used to check if input on the digital left button
  is detected
  */
  void readLeftButton();
  /*
  * This function is used to check if input on the digital middle button
  is detected
  */
  void readMiddleButton();
  /*
  * This function is used to check if input on the digital right button
  is detected
  */
  void readRightButton();

  /*
  * This function checks what setting has been selected and changes it's state
  * based on the selected setting
  */
  void selectSetting();

  /*
  * This function sets the rgb led to light up with a specific color
  * r = the red color of the led
  * g = the green color of the led
  * b = the blue color of the led
  */
  void setRGBLed(uint8_t r, uint8_t g, uint8_t b);
  /*
  * This function sets a tone to play for a specific input from 'A' to 'F'
  * input = the character to interpret a tone
  * duration = the duration of the tone
  */
  void setTone(char input, unsigned long duration);

  /*
  * This function checks the input of an user at each level of the game.
  * It takes user input, decodes it to a character, compares it with the
  * current character to guess and changes the game state based on the input
  */
  void readInput();
  /*
  * This function gets the specific character from the value sent on the analog
  * channel
  * value = the value sent on the analog channel
  */
  char decodeInput(int value);
  /*
  * This function generates an array of chars with the size based on the current
  * level. It considers the temperature and light sensors when generating this
  * sesquence
  */
  void generateSequence();
  /*
  * This fucntion repeats the sequence currently stored in the game memory.
  */
  void repeatSequence();
  /*
  * This function generates a random letter based on the temperature and 
  * light sensors
  * chars = buffer where it stores the character generated
  */
  char randLetter(char* chars);

  /*
  * This function prints a log message that constains the time and a specific
  * message on serial and on sd card.
  * message = specific message to be logged
  */
  void printLogMessage(char *message);

  /*
  * This function configures the timer that will be used as an internal counter
  */
  void configure_timer1();
  /*
  * This function initializes the timer that will be used as an internal counter
  */
  void init_timer1();

  GameState           _state;                                   // the current state of the game
  LiquidCrystal_I2C  *_lcd;                                     // the i2c lcd that shows data
  DS3231             *_rtc;                                     // the rtc module that stores time indefinetely
  Time               *_time;                                    // time structure to store time data
  DS18B20            *_ds;                                      // temperature sensor to store temperature data

  int                 _buttonState[3];                          // the current reading from the input pin
  int                 _lastButtonState[3]   = {LOW, LOW, LOW};  // the previous reading from the input pin
  unsigned long       _lastDebounceTime[3]  = {0, 0, 0};        // the last time the output pin was toggled
  unsigned long       _debounceDelay[3]     = {50, 50, 50};     // the debounce time; increase if the output flickers

  uint8_t             _currSetting = 0;                         // the current setting selected
  uint8_t             _currRetryOption = 0;                     // the current retry option selected

  uint8_t             _greenValue;                              // the current green value for the rgb led
  
  char                _buffer[10] = {'0', '0', '0', '0', '0', '0', '0', '0', '0', '0'}; // sequence of chars
  int                 _score;                                   // the total score obtained in a game session
  int                 _highScore;                               // the best score obtained during more game sessions
  int                 _level;                                   // the current level
  uint8_t             _lives;                                   // the total amount of lives
  uint8_t             _lettersRead;                             // total characters that the user took as input
  unsigned long       _keyPressed;                              // total amount of time a key was hold
};

#endif